"""
$Id: ReportDataTool.py,v 1.30 2002/05/28 23:53:07 evonnied Exp $
"""
##############################################################################
## copyright (C) 2002 Eugene von Niederhausern
##
##
##    This file is part of niederworksBRIDGE.
##
##    niederworksBRIDGE is free software; you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation; either version 2 of the License, or
##    (at your option) any later version.
##
##    niederworksBRIDGE is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with niederworksBRIDGE; if not, write to the Free Software
##    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
##############################################################################



__version__='$Revision: 1.30 $'[11:-2]

import sys,os,string
from Products.CMFCore.utils import UniqueObject, _checkPermission, getToolByName
from OFS.Folder import Folder
from Globals import InitializeClass, PersistentMapping, DTMLFile
from AccessControl import ClassSecurityInfo
from Acquisition import aq_base, aq_inner, aq_parent
from Products.CMFCore import CMFCorePermissions
from Products.ZSQLMethods.SQL import manage_addZSQLMethod
from string import join, split, replace, strip, lower
from Globals import package_home


_dtmldir = os.path.join( package_home( globals() ), 'dtml' )

class ReportDataTool (UniqueObject, Folder):
    '''
    This tool manages data connections and data format used for reports.
    '''
    id = 'portal_reportdata'
    meta_type = 'niederWorks Report Data Tool'
    
    security = ClassSecurityInfo()

    ##manage_options = ( { 'label' : 'Overview', 'action' : 'manage_overview' }
##                    , { 'label' : 'Data Connections', 'action' : 'manage_dataconnections' }
##                     , { 'label' : 'Output Formats'
##                       , 'action' : 'manage_outputformats'
##                       }
##                       ) + Folder.manage_options

    manage_options = ( { 'label' : 'Overview', 'action' : 'manage_overview' },
                       ) + Folder.manage_options
    sql_arguments = 'sql_fields sql_tables sql_where sql_groupby sql_having sql_orderby'
    sql_template = """ SELECT <dtml-var sql_fields> FROM <dtml-var sql_tables>
                        <dtml-if expr="_.len(sql_where) > 0">
                             WHERE <dtml-var sql_where>
                        </dtml-if>
                        <dtml-if expr="_.len(sql_groupby)">
                            GROUP BY <dtml-var sql_groupby>
                        </dtml-if>
                        <dtml-if expr="_.len(sql_having)">
                            HAVING <dtml-var sql_having>
                        </dtml-if>
                        <dtml-if expr="_.len(sql_orderby)">
                        ORDER BY <dtml-var sql_orderby>
                        </dtml-if>
                        """
    
    #
    #   ZMI methods
    #
    security.declareProtected( CMFCorePermissions.ManagePortal
                             , 'manage_overview' )
    manage_overview = DTMLFile( 'explainReportDataTool', _dtmldir )

    #manage_dataconnections = DTMLFile('manage_dataconnections',_dtmldir)
    #manage_outputformats = DTMLFile('manage_outputformats',_dtmldir)

    _operator_map = {'generic':{'text':['=','!=','Like','Not Like','is Null','is not Null']
					 ,'number':['=','!=','<','<=','>','>=', 'is Null','is not Null','IN']
					 ,'date':['=','!=','Like','Not Like','is Null','is not Null']
					 }
					  ,'pgsql':{'text':['=','!=','~','!~','Like','Not Like','is Null','is not Null']
					 ,'number':['=','!=','<','<=','>','>=', 'is Null','is not Null']
					 ,'date':['=','!=','Like','Not Like','is Null','is not Null']
					 }
					  }
    _orderby_map={'ascending':'ASC','descending':'DESC'}
    _agg_func = {'count':'count(%s)','count distinct':'count( distinct %s)','sum':'sum(%s)','average':'avg(%s)','maximum':'max(%s)','minimum':'min(%s)','std dev':'std(%s)'}
    security.declareProtected( CMFCorePermissions.ManagePortal
                             , 'manage_addDataConnectionForm')
    def listDataConnections(self):
        results = []
        for obj in self.objectValues():
            if hasattr(obj,'_isAnSQLConnection') and obj._isAnSQLConnection:
                results.append(obj)
        return results

    def getDataConnection(self,connect_id):
        for obj in self.listDataConnections():
            if connect_id == obj.getId():
                return obj
        return None

    def getTableList(self,connect_id):
        con = self.getDataConnection(connect_id)
        browser = getattr(con,'manage_get_tables',None)
        o=[]
        if browser:
            tables = browser()
            for t in tables:
                if t.has_key('tablename'):
                    o.append(t['tablename'])
            

        return o
            
    def getTableInfo(self,connect_id,tablename):
        con = self.getDataConnection(connect_id)
        browser = getattr(con,'manage_get_table_infos',None)
        o = []
        if browser:
            table_info = browser(tablename)
            for ti in table_info:
                if ti.has_key('attname'):
                    o.append(ti['attname'])

        return o
                    
    def getFieldDataType(self,connect_id,tablename,fieldname):
        con = self.getDataConnection(connect_id)
        browser = getattr(con,'manage_get_table_infos',None)
        t = ''
        if browser:
            table_info = browser(tablename)
            for ti in table_info:
                if ti.has_key('attname') and ti['attname'] == fieldname:
                    if ti['typname'] in ['int','int8','int4','float','float4','float8']:
                        t = 'number'
                    elif ti['typname'] in ['date','timestamp','time']:
                        t = 'date'
                    else:
                        t = 'text'

        return t

                             
    
    def getZSqlMethod(self,connect_id):
        zsql_id ='%s_zsqlmethod' % (connect_id )
        for obj in self.objectValues(['Z SQL Method']):
            if obj.getId() == zsql_id :
                return obj

        manage_addZSQLMethod(self,zsql_id,'',connect_id,self.sql_arguments,self.sql_template)
        return self.getZSqlMethod(connect_id)
        
        
    
    def listOutputTemplates(self):

        return  self.objectValues(['DTML Method','Script (Python)','Page Template'])

            

    def getOutputTemplate(self,method_id='default'):
        for obj in self.listOutputTemplates():
            if method_id == obj.getId():
                return obj
        return None
    def reportColumns(self,userreport):
        o = []
        reporttable=userreport.getReportTable()
        for field in userreport.userReportFields():
            rtf = reporttable.getReportTableField(field.reportTableFieldId())
            if (field.fieldAlias() == rtf.fieldName() and not rtf.isAggregate()) or (not field.fieldAlias() and not rtf.isAggregate()):                
                f_name = rtf.fieldName()
            elif rtf.isAggregate() and not field.fieldAlias():
                f_name = '%s' % ( rtf.fieldName() )
            else:
                f_name = '%s' % ( field.fieldAlias())
                
            if field.isVisible() and rtf.isVisible():
                if rtf.isAggregate():
                    if not field.fieldAlias():
                        f_name = rtf.fieldName()
                o.append(f_name)
        return o
    
    def _buildFields(self,userreport,reporttable):
        output = ""
        for field in userreport.userReportFields():
            rtf = reporttable.getReportTableField(field.reportTableFieldId())
            if not rtf.isAggregate() and ((field.fieldAlias() == rtf.fieldName()) or not field.fieldAlias()):
                alias_part = ''
            elif rtf.isAggregate() and not field.fieldAlias():
                alias_part = ' as "%s" ' % ( rtf.fieldName() )
            else:
                if not field.fieldAlias() and rtf.isAggregate():
                    alias_part = ' as "%s" '% ( rtf.fieldName() )
                else:
                    alias_part = ' as "%s" ' % ( field.fieldAlias())
                
            if field.isVisible() and rtf.isVisible():
                if rtf.isAggregate():
                    if len(output) == 0:
                        output = ' %s %s ' % ( rtf.aggregateFunc() % rtf.fieldName(),alias_part )
                    else:
                        output = ' %s, %s %s ' % ( output,rtf.aggregateFunc() % rtf.fieldName(),alias_part )
                else:
                    if len(output) == 0:
                        output = ' %s %s ' % ( rtf.fieldName(),alias_part )
                    else:
                        output = ' %s, %s %s' % ( output,rtf.fieldName(),alias_part )


                        
        return output
    def _buildTables(self,userreport, reporttable):
        output = ' %s ' % (reporttable.table())
        return output
    
    def _buildWhere(self,userreport,reporttable,rtinput=None,isAggregate=0):
        where = ""

        if not rtinput: rtinput={}
        
        for field in userreport.userReportFields():
            rtf = reporttable.getReportTableField(field.reportTableFieldId())
            if rtf.isSearchable() and (field.fieldValue() or rtinput.has_key(field.getId())) and rtf.isAggregate() == isAggregate:
                
                if field.fieldValue() :
                    field_value = field.fieldValue()
                    

                if rtf.defaultOperator() :
                    operator = rtf.defaultOperator()
                    
                if field.operator():
                    operator = field.operator()
                    
                if rtinput.has_key(field.getId()):
                    if rtinput[field.getId()].has_key('field_value'):
                        field_value = rtinput[field.getId()]['field_value']
                    if rtinput[field.getId()].has_key('operator'):
                        operator=rtinput[field.getId()]['operator']

                if string.lower(operator).find('null') >= 0:
                    field_value = None
                
                if len(where) == 0:
                    if rtf.isAggregate():
                        where = ' %s %s %s ' % (rtf.aggregateFunc() % rtf.fieldName(),operator,self._quote(rtf.dataType(),field_value))
                    else:
                        where = ' %s %s %s ' % (rtf.fieldName(),operator,self._quote(rtf.dataType(),field_value))
                else:
                    if rtf.isAggregate():
                        where = ' %s AND %s %s %s ' % (where,rtf.aggregateFunc() % rtf.fieldName(),operator,self._quote(rtf.dataType(),field_value))
                    else:
                        where = ' %s AND %s %s %s ' % (where,rtf.fieldName(),operator,self._quote(rtf.dataType(),field_value))

                    
        for field in reporttable.reportTableFields():
            if not field.isVisible() and field.defaultValue()  and field.isAggregate() == isAggregate and field.defaultOperator():

                field_value = field.defaultValue()
                
                if string.lower(field.defaultOperator()).find('null') >= 0:
                    field_value = None
                
                if len(where) == 0:
                    if field.isAggregate():
                        where = ' %s %s %s ' % (field.aggregateFunc() % field.fieldName(),field.defaultOperator(),self._quote(field.dataType(),field_value))
                    else:
                        where = ' %s %s %s ' % (field.fieldName(),field.defaultOperator(),self._quote(field.dataType(),field_value))
                else:
                    if field.isAggregate():
                        where = '%s AND  %s %s %s ' % (where,field.aggregateFunc() % field.fieldName(),field.defaultOperator(),self._quote(field.dataType(),field_value))
                    else:
                        where = '%s AND  %s %s %s ' % (where,field.fieldName(),field.defaultOperator(),self._quote(field.dataType(),field_value))
        return where
    
    def _buildHaving(self,userreport,reporttable,rtinput=None):
        return self._buildWhere(userreport = userreport,reporttable = reporttable,rtinput = rtinput,isAggregate = 1)
    
    def _buildGroupBy(self,userreport,reporttable):
        output = ""
        if reporttable.hasAggregate():
            
            for field in userreport.userReportFields():
                rtf = reporttable.getReportTableField(field.reportTableFieldId())
                if not rtf.aggregateFunc() and field.isVisible():
                    if len(output) == 0:
                        output = ' %s ' % (rtf.fieldName())                    
                    else:
                        output = ' %s , %s ' % (output,rtf.fieldName())                    
        return output

    def _buildOrderBy(self,userreport,reporttable):
        output = ""
        for field in userreport.userReportFields():
            rtf = reporttable.getReportTableField(field.reportTableFieldId())
            orderby = self.orderByValue(field.orderBy())
            if rtf.isSortable() and orderby :
                if rtf.isAggregate():
                    fieldname = field.fieldAlias()
                else:
                    fieldname = rtf.fieldName()
                    
                if len(output) == 0:
                    output = ' %s %s ' % (fieldname,orderby)                    
                else:
                    output = ' %s , %s %s ' % (output,fieldname,orderby)
                    
        return output
    
    
    def _buildQuery(self,userreport,reporttable,rtinput=None):
        """ to build the query:
            * get user report
            * get report table
            * find connection object
            * build fields
            * build tables
            * build WHERE
            * build group by
            * build having
            * build order by
        """
        sql_fields = self._buildFields(userreport,reporttable)
        sql_tables = self._buildTables(userreport,reporttable)
        sql_where = self._buildWhere(userreport,reporttable,rtinput)
        sql_group = self._buildGroupBy(userreport,reporttable)
        sql_having = self._buildHaving(userreport,reporttable,rtinput)
        sql_order = self._buildOrderBy(userreport,reporttable)
        
        

        return {'sql_fields':sql_fields,'sql_tables':sql_tables,'sql_where':sql_where
                ,'sql_groupby':sql_group,'sql_having':sql_having,'sql_orderby':sql_order}
        
    def _executeQuery(self,userreport_id,rtinput=None):
        userreport = self.getUserReport(userreport_id)
        reporttable = self.getReportTable(userreport.reportTableId())
        
        query = self._buildQuery(userreport,reporttable,rtinput)
        sql_method = self.getZSqlMethod(reporttable.connectionId())
        if not sql_method:
            raise 'EXECUTE_FAILED',''
        res = sql_method(query)
        return res
    def getSQL(self,userreport_id,rtinput=None):
        userreport = self.getUserReport(userreport_id)
        reporttable = self.getReportTable(userreport.reportTableId())
        
        query = self._buildQuery(userreport,reporttable,rtinput)
        sql = 'SELECT %(sql_fields)s FROM %(sql_tables)s ' % query
        if len(query['sql_where'])>0 :
            sql = sql + ' WHERE %(sql_where)s ' % query
        if len(query['sql_groupby'])>0 :
            sql = sql + ' GROUP BY %(sql_groupby)s ' % query
        if len(query['sql_having']) > 0 :
            sql = sql + ' HAVING %(sql_having)s ' % query
        if len(query['sql_orderby']) > 0 :
            sql = sql + ' ORDER BY %(sql_orderby)s ' % query

        return sql
    
    def executeQuery(self,userreport_id,rtinput=None):
        res = self._executeQuery(userreport_id,rtinput)
        return res
    
    def outputQuery(self,userreport_id,output_template='default',rtinput=None,REQUEST=None):
        res = self.executeQuery(userreport_id,rtinput)
        tmpl = self.getOutputTemplate(output_template)
        
        output = tmpl(REQUEST=REQUEST,results=res)
        return output
    
    def executeQuery_XMLRPC(self,userreport_id,output_template='xmlrpc',rtinput=None,REQUEST=None):
        return self.executeQuery(userreport_id,output_template,rtinput,REQUEST)
    
    def getUserReport(self,userreport_path):
        cat_tool = getToolByName(self,'portal_catalog')
        obj = None    
        if len(userreport_path) > 0:
            path = userreport_path[0:string.rfind(userreport_path,'/')]
            id =  userreport_path[string.rfind(userreport_path,'/') + 1:]
            
            res = cat_tool.searchResults({'meta_type':'niederWorks User Report','path':path,'id':id})
            if len(res) >= 1:
                obj = res[0].getObject()
        return obj
    
    def getReportTable(self,reporttable_path):
        cat_tool = getToolByName(self,'portal_catalog')
        obj = None    
        if len(reporttable_path) > 0:
            path = reporttable_path[0:string.rfind(reporttable_path,'/')]
            id =  reporttable_path[string.rfind(reporttable_path,'/') + 1:]
            
            res = cat_tool.searchResults({'meta_type':'niederWorks Report Table','path':path,'id':id})
            if len(res) >= 1:
                obj = res[0].getObject()
        return obj
    
    def listReportTables(self,rv_state='published'):
        results = []
        cat_tool = getToolByName(self,'portal_catalog')
        res = cat_tool.searchResults({'meta_type':'niederWorks Report Table'})
        for r in res:
            #change this to list report tables based on user permissions...
            if r.review_state == rv_state:
                obj = r.getObject()
                results.append(obj)
        return results

    
    def _quote(self,dtype,value):
        if  value is None:
            return ''
        
        if dtype in ('text','date'):
            return " '%s' " % (value)
        elif dtype in ('number',):
            return " %s " % (value)
        else:
            return " %s " % (value)
        
    def operatorList(self,datatype='text',db_type='generic'):
		
        if self._operator_map.has_key(db_type):
            if self._operator_map[db_type].has_key(datatype):
                return self._operator_map[db_type][datatype]
        return []
    
    def dataTypeList(self,db_type='generic'):
		
        if self._operator_map.has_key(db_type):
            return self._operator_map[db_type].keys()
        return []

    def aggregateFunctionList(self):
        return self._agg_func.keys()

    def aggregateFunction(self,func_name):
        if self._agg_func.has_key(func_name):
            return self._agg_func[func_name]
        return ''
    
    def orderByList(self):
        return  ['None'] + self._orderby_map.keys() 
    
    def orderByValue(self,name='None'):
        if self._orderby_map.has_key(name):
            return self._orderby_map[name]
        return ''
    
    def valueList(self,id):
        if id in self.objectIds():
            return self[id]()
        return None
	
    def listValueLists(self):
        r=[]
        for obj in self.objectValues(['DTML Method','Script (Python)','Page Template']):
            r.append(obj)
        return r
        
