"""
$Id: UserReport.py,v 1.25 2002/05/29 03:26:59 evonnied Exp $
"""
##############################################################################
## copyright (C) 2002 Eugene von Niederhausern
##
##
##    This file is part of niederworksBRIDGE.
##
##    niederworksBRIDGE is free software; you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation; either version 2 of the License, or
##    (at your option) any later version.
##
##    niederworksBRIDGE is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with niederworksBRIDGE; if not, write to the Free Software
##    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
##############################################################################

import string
from Globals import Persistent, PersistentMapping,InitializeClass

from AccessControl import ClassSecurityInfo

from Acquisition import Implicit, aq_base,aq_inner, aq_parent

from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.utils import getToolByName
from Products.CMFDefault.DublinCore import DefaultDublinCoreImpl

from Products.CMFCore.PortalContent import PortalContent

from OFS.SimpleItem import SimpleItem

factory_type_information = ( {'id'             : 'User Report'
                              , 'meta_type'    : 'niederWorks User Report'
                              , 'description'  : 'User Report is a way to view data from a database using report table templates'
                              , 'icon'         : 'userreport_icon.gif'
                              , 'product'      : 'niederworksBRIDGE'
                              , 'factory'      : 'addUserReport'
                              , 'actions'        :
                                    ( { 'id'           : 'view'
                                        ,'name'        : 'View'
                                        ,'action'      : 'userreport_view'
                                        ,'permissions' : ( CMFCorePermissions.View,)
                                       }
                                      , { 'id'         : 'edit'
                                          ,'name'      : 'Edit'
                                          ,'action'    : 'userreport_edit_form'
                                          ,'permissions':( CMFCorePermissions.ModifyPortalContent,)
                                        }
                                      ,),
                              },)

def addUserReport(self
                   ,id
                   ,reporttable_id=''
                   ,title = ''
                   ,description = ''
                   , REQUEST = None):
    userreport = UserReport(id,reporttable_id,title,description)
    self._setObject(id,userreport)

class UserReport(PortalContent,Persistent,Implicit,DefaultDublinCoreImpl):
    meta_type = 'niederWorks User Report'

    security = ClassSecurityInfo()

    def __init__(self
                 , id
                 , reporttable_id = ''
                 , title = ''
                 , description = ''
                 ):
        DefaultDublinCoreImpl.__init__(self)
        self.id = id
        self._reporttable_id = reporttable_id
        self.setTitle(title)
        self.setDescription(description)
        self._fields = PersistentMapping()
    security.declareProtected(CMFCorePermissions.ModifyPortalContent,'edit')
    def edit(self, title = ''
                 , description = ''                 
                 , reporttable_id = ''
                 ):
        if  reporttable_id and self._reporttable_id != reporttable_id:
            self._fields = PersistentMapping()
            
            self._reporttable_id = reporttable_id

        self.setTitle(title)
        self.setDescription(description)
        
    security.declarePublic('reportTableId')
    def reportTableId(self):
        return self._reporttable_id
    
    security.declarePublic('getReportTable')
    def getReportTable(self):
        rd_tool = getToolByName(self,'portal_reportdata')
        rt = rd_tool.getReportTable(self.reportTableId())
        return rt

    security.declarePublic('getReportTableField')
    def getReportTableField(self,rtf_id):
        rt = self.getReportTable()
        rtf = rt.getReportTableField(rtf_id)
        return rtf

    security.declarePublic('userReportFields')        
    def userReportFields(self):
        fields = self._fields.keys()
        fields.sort()
        o=[]
        for f in fields:
            o.append(self._fields[f])
        return o

    security.declarePublic('userReportFieldNames')
    def userReportFieldNames(self):
        return self._fields.keys().sort()
    
    security.declarePublic('getUserReportField')    
    def getUserReportField(self,field_name):
        if self._fields.has_key(field_name):
            field = self._fields[field_name]
            return field.__of__(self)

        return None
    def __getattr__(self,name):
        if self._fields.has_key(name):
            return self.getUserReportField(name)
        raise AttributeError,'%s not found' % (name)
    security.declareProtected(CMFCorePermissions.ModifyPortalContent,'addUserReportField')
    def addUserReportField(self
                            , id
                            , reporttable_field_id
                            , title = ''
                            , operator = ''
                            , field_value = ''
                            , orderby = ''
                            , visible = 1
                            ):
        if self._fields.has_key(id) :
            raise 'FIELD_EXISTS','Field already Exists %s' % (id)
        
        rtf = UserReportField(
            id
            , reporttable_field_id
            , title
            , operator
            , orderby
            , visible
            )
        self._fields[id] = rtf
        return rtf
    security.declareProtected(CMFCorePermissions.ModifyPortalContent,'deleteUserReportField')    
    def deleteUserReportField(self,id):
        del self._fields[id]
    security.declarePublic('hasRunTimeInput')        
    def hasRunTimeInput(self):
        for urf in self.userReportFields():
            if urf.runTimeInput():
                return 1
        return 0

    security.declarePublic('runTimeInputForm2')
    def runTimeInputFormHidden(self):
        rt = self.getRunTimeInput()
        output = ''
        if rt:
            for r in rt.keys():
                output = '%s <input type="hidden" name="report_fields.id:records" value="%s">' % (output,r)
                output = '%s <input type="hidden" name="report_fields.operator:records" value="%s">' % (output,rt[r]['operator'])

                output = '%s <input type="hidden" name="report_fields.field_value:records" value="%s">' % (output,rt[r]['field_value'])
        return output
    security.declarePublic('runTimeInputForm')
    def runTimeInputForm(self):
        x=[]
        rdata_tool = getToolByName(self,'portal_reportdata')
        rt = rdata_tool.getReportTable(self.reportTableId())
        output=''
        for urf in self.userReportFields():
            rtf = rt.getReportTableField(urf.reportTableFieldId())
            x.append((urf,rtf))
        if len(x)> 0:
            output = '<table>'
            for (urf,rtf) in x:
                output = ' %s <tr><th>%s</th> ' % (output,urf.Title())
                if urf.runTimeInput():
                    output = ' %s <td> <select name="report_fields.operator:records" >  ' % (output)
                    output = ' %s <option value=''>(select value)</option>' % (output)
                    for o in rdata_tool.operatorList(rtf.dataType(),rt.dbType()):
                        if o == urf.operator():
                            output = ' %s <option selected value="%s">%s</option>' % (output,o,o)
                        else:
                            output = ' %s <option value="%s">%s</option>' % (output,o,o)
                    output = ' %s </select></td> ' % (output)

                    output = '%s<input type="hidden" name="report_fields.id:records" value="%s"></td><td>' % (output,urf.getId())

                    if rtf.valueList():
                        list = rdata_tool.valueList(rtf.valueList())
                        if list:
                            output = ' %s <select name="report_fields.field_value:records" > ' % (output)
                            output = '%s <option value=''>(select value)</option>' % (output)
                            for item in list:
                                if urf.fieldValue() == item:

                                    output = '%s <option selected value="%s">%s</option>' % (output,item,item)
                                else:
                                    output = '%s <option value="%s">%s</option>' % (output,item,item)
                            output = '%s </select>' % ( output )
                    else:
                        output = ' %s <input type="text" name="report_fields.field_value:records" ></td> ' % (output)

                else:
                    if not urf.operator():
                        op = ''
                    else:
                        op = urf.operator()
                        
                    if not urf.fieldValue():
                        fv = ''
                    else:
                        fv = urf.fieldValue()
                        
                    output = ' %s <td>%s</td><td>%s</td> ' % (output,op,fv)
                    
                output='%s </tr>' % (output)
            output = '%s </table>' % (output)
        return output
    security.declarePublic('getRunTimeInput')
    def getRunTimeInput(self,REQUEST=None):
        rtinput = None
        if not REQUEST: REQUEST=getattr(self,'REQUEST',None)
        if REQUEST and REQUEST.has_key('report_fields'):
            report_fields = REQUEST.report_fields
            if len(report_fields)>0:
                rtinput={}
                for field in report_fields:
                    if not getattr(field,'field_value',None) is None and len(field.field_value)>0  :

                        rtinput[field.id]={}
                        rtinput[field.id]['field_value']=field.field_value
                        if field.operator is not None and len(field.operator) > 0 :
                            rtinput[field.id]['operator'] = field.operator
        return rtinput
    
    # for debugging
    security.declarePublic('getSQL')
    def getSQL(self,REQUEST=None):
        rtinput = self.getRunTimeInput()

        
        rd_tool = getToolByName(self,'portal_reportdata')

        return rd_tool.getSQL(userreport_id = string.join(self.getPhysicalPath(),'/'),rtinput = rtinput)

    security.declarePublic('reportColumns')
    def reportColumns(self):
        rd_tool = getToolByName(self,'portal_reportdata')
        return rd_tool.reportColumns(self)
    
    security.declarePublic('results')
    
    def results(self,REQUEST=None):
        rtinput = self.getRunTimeInput()
                        
        rd_tool = getToolByName(self,'portal_reportdata')

        res = rd_tool.executeQuery(userreport_id = string.join(self.getPhysicalPath(),'/'),rtinput = rtinput)
        return res
        
InitializeClass(UserReport)

class UserReportField(Persistent,Implicit,SimpleItem):
    meta_type = 'niederWorks User Report Field'
    
    security = ClassSecurityInfo()
    
    def __init__(self
                 , id                 
                 , reporttable_field_id
                 , title = ''
                 , operator = ''
                 , field_value = ''
                 , orderby = ''
                 , visible = 1
                 , runtime_input = 0
                 ):
        self.id = id
        self.title  = title
        self._reporttable_field_id = reporttable_field_id
        self._operator = operator
        self._field_value = field_value
        self._orderby = orderby
        self._visible = visible
        self._runtime_input = runtime_input

    security.declarePublic('Title')
    def Title(self):
        return self.title
    
    security.declareProtected(CMFCorePermissions.ModifyPortalContent,'edit')
    def edit(self
                 , title = ''
                 , operator = ''
                 , field_value = ''
                 , orderby = ''
                 , visible = 1
                 , runtime_input = 0
                 ):
        self.title = title
        self._operator = operator
        self._field_value = field_value
        self._orderby = orderby
        self._visible = visible
        self._runtime_input = runtime_input
    security.declarePublic('isVisible')
    def isVisible(self):
        return self._visible
    
    security.declarePublic('runTimeInput')
    def runTimeInput(self):
        return self._runtime_input
    
    security.declarePublic('fieldAlias')
    def fieldAlias(self):
        return self.title

    security.declarePublic('reportTableFieldId')
    def reportTableFieldId(self):
        return self._reporttable_field_id

    

    security.declarePublic('operator')
    def operator(self):
        if self._operator and  len(self._operator) > 0:
            return self._operator
        return None

    security.declarePublic('fieldValue')
    def fieldValue(self):
        if self._field_value and len(self._field_value) > 0  :
            return self._field_value
        return None
    
    security.declarePublic('orderBy')
    def orderBy(self):
        if self._orderby in ('ascending','descending') :
            return self._orderby
        return ''
                 

        
InitializeClass(UserReportField)
